      Program driver_gmcut
c     ------------------------------------------------------------------
c
c     driver_gmcut:  Driver for gmcut, a GRASP for finding approximate
c                    solutions to the Maximum Cut Problem 
c
c
c     authors:       Paola Festa             [paofes@udsab.dia.unisa.it]
c                    Mauricio G.C. Resende   [mgcr@research.att.com]
c                    Panos P.M. Pardalos     [pardalos@ufl.edu]
c                    Celso C. Ribeiro        [celso@inf.puc-rio.br]
c
c     ------------------------------------------------------------------
c
c     Dimensions of maximum graph size:
c
c          maxe  - max number of arcs 
c          maxv  - max number of vertices
c
c     ------------------------------------------------------------------
      implicit none
      integer    maxe,maxv,maxelt
      parameter  (maxe=300000,maxv=30000,maxelt=50)
c     ------------------------------------------------------------------
c
c     Other array dimension.
c
c          max2e  - twice max number of edges
c          max2v  - twice max number of vertices
c
c     ------------------------------------------------------------------
      integer    max2e,max2v
      parameter  (max2e=2*maxe,max2v=2*maxv)    

c     ------------------------------------------------------------------
c
c     Input/output device numbers passed to errmsg, gfvs, outvtx, and
c     readp: 
c
c          in     - Fortran input device
c          out    - Fortran output device
c
c     ------------------------------------------------------------------
      integer   in,out
      parameter (in=5,out=6)

c     ------------------------------------------------------------------
c
c     Scalars passed to gmcut:
c
c          alpha  - GRASP RCL parameter
c          bestz  - sum of the weights of the arcs of best cutset found
c          bnS
c          errcnd - error condition
c          gitrb  - iteration cut best set was found
c          look4  - size of cutset sought
c          maxitr - maximum number of GRASP iterations
c          kmax   - maximum k for VNS
c          ne     - number of arcs in original graph
c          nmxelt - 
c          nv     - number of nodes in original graph
c          pron   - indicator of path relinking (on/off)
c          prttyp - iteration summary report
c                   = 0 - silent
c                   = 1 - show improvements only
c                   = 2 - show all iterations
c          seedb  - seed at start of best iteration
c          seedd  - seed for pseudo random number generator
c          seedp  - seed for path relinking
c          seedn  - seed for VNS
c
c     ------------------------------------------------------------------
      real             alpha,hammd
      integer          bestz,bnS,errcnd,gitrb,look4,maxitr,
     +                 ne,n2e,nmxelt,nv,pron,prttyp,seedb,nelt,seedp,
     +                 prtype,i,objf,poston,seedn,kmax,n2v,nmelt
      double precision seedd

c     ------------------------------------------------------------------
c
c     Arrays passed to gmcut:
c           
c           bS      - best set S found          
c           deg     - in-degree of vertex (copy)
c           deg0    - in-degree of vertex
c           degs
c           degsb
c           edge
c           edge0
c           felt 
c           first   - pointer to start of in-adjacency list (copy)
c           first0  - pointer to start of in-adjacency list (original) 
c           next    - pointer to next element of in-adjacency 
c                     list (copy)
c           next0   - pointer to next element of in-adjacency 
c                     list (original)
c           nxtelt
c           pout    - pointer to start of out-adjacency list (copy)
c           pout0   - pointer to start of out-adjacency list (original)
c           rcl     - Restricted Candidate List
c           S       - set S
c           Sbar    - set Sbar
c           sums
c           sumst
c           sumsb
c           sumsbt
c           v       - vertex in in-adjacency list (copy)
c           v0      - vertex in in-adjacency list (original)
c           velt
c           vertex  - 
c           vtx1    - vertex 1 of arc
c           vtx2    - vertex 2 of arc
c           w       - arc weights array (original)
c           ws      - sorted arc weights array 
c           wsind   - indices of sorted arc weights array 

c     ------------------------------------------------------------------
      integer    bS(maxv),deg(maxv),deg0(maxv),tvnS(maxv),tvnSb(maxv),
     +           degs(maxv),degsb(maxv),edge(max2e),edge0(max2e),
     +           felt(maxelt),
     +           first(maxv),first0(maxv),
     +           next(max2e),next0(max2e),
     +           nxtelt(maxelt*maxv),objelt(maxelt),
     +           rcl(max2v),S(maxv),optS(maxv),optbS(maxv),tmpS(maxv),
     +           Sbar(maxv),sums(maxv),sumst(maxv),sumsb(maxv),
     +           sumsbt(maxv),symdif(maxv),
     +           v(max2e),v0(max2e),velt(maxelt*maxv),
     +           vertex(maxv),
     +           vtx1(maxe),vtx2(maxe),w(maxe),
     +           ws(maxe),wsind(maxe),x(maxv),y(maxv),tmpx(maxv),
     +           tmpy(maxv),x0(maxv),xdiff(maxv),ydiff(maxv),
     +           xn(maxv),nindex(maxv)

c     ------------------------------------------------------------------
c     Sets default values for parameters.
c     ------------------------------------------------------------------

      call dflset(alpha,hammd,look4,maxitr,nmxelt,pron,prttyp,seedd,
     +            seedp) 
      maxitr=1000
      prttyp=1
      hammd=.01
      nmxelt=30
      nelt=0
      pron=0
      alpha=0.
      poston=0
      prtype=1
      seedn=270001
      kmax=100

      read *,seedd,seedp,look4

c     ------------------------------------------------------------------
c     Read input graph.
c     ------------------------------------------------------------------
      call readp(errcnd,in,ne,nv,maxe,maxv,vtx1,vtx2,w)
      n2v = 2*nv
      n2e = 2*ne
      nmelt=nv*nmxelt
c     ------------------------------------------------------------------
c
c     Algorithm parameters:
c
c     alpha  -   GRASP parameter alpha (real)
c
c     look4  -   GRASP returns if cutset set of size at least equal
c                to look4 less or equal to nv is found (integer)
c
c     maxitr -   maximum number of GRASP iterations (integer)
c
c     prttyp -   indicator of type of iteration report (integer)
c                * silent -- prttyp = 0
c                * show only improvements -- prttyp = 1
c                * show all iterations -- prttyp = 2
c
c     seed   -   seed for random number generator in [0,2^{31}-1]
c                (integer) 
c
c     ------------------------------------------------------------------
c     Solve problem approximately using GRASP.
c     ------------------------------------------------------------------
      if (errcnd.eq.0) then
         call gmcut(alpha,bS,bestz,bnS,
     +                 deg,deg0,degs,degsb,edge,edge0,
     +                 errcnd,felt,
     +                 first,first0,gitrb,hammd,
     +                 look4,
     +                 maxe,maxelt,maxitr,maxv,max2e,max2v,ne,n2e,nelt,
     +                 next,next0,nmxelt,nmelt,
     +                 nv,n2v,nxtelt,objelt,
     +                 out,pron,prttyp,rcl,
     +                 S,Sbar,seedb,seedd,seedp,sums,sumst,sumsb,
     +                 sumsbt,symdif,tmpx,tmpy,v,v0,velt,
     +                 vertex,
     +                 vtx1,vtx2,w,ws,wsind,x,y,optS,tmpS,prtype,x0,
     +                 xdiff,ydiff,poston,xn,nindex,kmax,seedn,tvnS,
     +                 tvnSb)

         call errmsg(errcnd,out)
c        ---------------------------------------------------------------
c        Print the cut set.
c        ---------------------------------------------------------------
         if (errcnd .eq. 0) then
              call outsol(bestz,bS,gitrb,bnS,nv,out,seedb,vertex,
     +                    ne,vtx1,vtx2,w)
         endif
      else
         call errmsg(errcnd,out)
      endif
      print *,' errcnd = ',errcnd

c     ------------------------------------------------------------------
      stop
c     ------------------------------------------------------------------
c     End driver for gmcut.f: A GRASP for the Max-Cut Problem
c     ------------------------------------------------------------------

      end
