      subroutine gmcut(alpha,bS,bestz,bnS,
     +                 deg,deg0,degs,degsb,edge,edge0,
     +                 errcnd,felt,
     +                 first,first0,gitrb,hammd,
     +                 look4,
     +                 maxe,maxelt,maxitr,maxv,max2e,max2v,ne,n2e,nelt,
     +                 next,next0,nmxelt,nmelt,
     +                 nv,n2v,nxtelt,objelt,
     +                 out,pron,prttyp,rcl,
     +                 S,Sbar,seedb,seedd,seedp,sums,sumst,sumsb,
     +                 sumsbt,symdif,tmpx,tmpy,v,v0,velt,
     +                 vertex,
     +                 vtx1,vtx2,w,ws,wsind,x,y,optS,tmpS,prtype,x0,
     +                 xdiff,ydiff,poston,xn,nindex,kmax,seedn,tvnS,
     +                 tvnSb)

c     ------------------------------------------------------------------
c
c     gmcut: A package of FORTRAN subroutines for finding a max-cut
c            set in a graph using GRASP.  The set is not guaranteed to
c            be optimal.
c
c     authors:       Paola Festa             [paofes@udsab.dia.unisa.it]
c                    Mauricio G.C. Resende   [mgcr@research.att.com]
c                    Panos P.M. Pardalos     [pardalos@ufl.edu]
c
c     ------------------------------------------------------------------
c
c     Passed input scalars:
c
c          alpha  - GRASP RCL parameter
c          hammd  -
c          kmax   - maximum k for VNS
c          look4  - size of cutset sought
c          maxc   - upper bound on maxcut cost
c          max2e  - array dimension = 2*maxe
c          max2v  - array dimension = 2*maxv
c          maxe   - array dimension
c          maxitr - maximum number of GRASP iterations
c          maxv   - array dimension
c          ne     - number of arcs in original graph
c          n2e    - array dimension = 2*ne
c          n2v    - array dimension = 2*nv
c          nv     - number of vertices in original graph 
c          out    - Fortran output device
c          prttyp - iteration summary report
c                   = 0 - silent
c                   = 1 - show improvements only
c                   = 2 - show all iterations
c
c     ------------------------------------------------------------------
      implicit   none
      real       alpha,hammd
      integer    look4,maxc,max2e,max2v,maxe,maxelt,maxitr,maxv,
     +           mindif,ne,n2e,n2v,nv,out,prttyp,elitex,prtype,kmax

c     ------------------------------------------------------------------
c
c     Passed input/output scalars:
c
c          errcnd - error condition
c          nS
c          nSbar 
c          seed   - seed for pseudo random number generator
c          seedd  - seed for pseudo random number generator
c          seedp  - seed for path relinking
c          seedn  - seed for VNS
c
c     ------------------------------------------------------------------
      integer          errcnd,nS,nSbar,seed,seedp,seedn,ntvnS,ntvnSb,
     +                 noptS
      double precision seedd

c     ------------------------------------------------------------------
c
c     Passed output scalars:
c
c          bestz  - sum of the arc weights of cutset found
c          bnS
c          gitrb  - number of iterations to find best cutset
c          nmxelt
c          seedb  - seed at start of best iteration 
c
c     ------------------------------------------------------------------
      integer    bestz,bnS,gitrb,nmxelt,nmelt,seedb

c     ------------------------------------------------------------------
c
c
c     Passed input/output arrays:
c
c           bS      -
c           S       -
c           Sbar    -
c           deg     - degree
c           deg0    -
c           degs
c           degsb
c           edge
c           edge0
c           felt
c           first   - pointer to start of adjacency list (copy)
c           first0  - pointer to start of adjacency list (original)
c           next    - pointer to next element of adjacency
c                     list (copy)
c           next0   - pointer to next element of adjacency
c                     list (original)
c           nxtelt
c           objelt
c           rcl     - Restricted Candidate List
c           sums
c           sumst
c           sumsb
c           sumsbt
c           vertex  -   
c           v       - vertex in adjacency list (copy)
c           v0      - vertex in adjacency list (original)
c           velt
c           vtx1    - vertex 1 of arc
c           vtx2    - vertex 2 of arc
c           w       - arc weights array (original)
c           ws      - sorted arc weights array
c           wsind   - edge indices of sorted weights array
c
c     ------------------------------------------------------------------
      integer    bS(nv),deg(nv),deg0(nv),
     +           degs(nv),degsb(nv),edge(n2e),edge0(n2e),
     +           felt(nmxelt),optS(nv),tmpS(nv),
     +           first(nv),first0(nv),
     +           next(n2e),next0(n2e),
     +           nxtelt(nmelt),objelt(nmxelt),
     +           rcl(n2v),S(nv),Sbar(nv),
     +           sums(nv),sumst(nv),
     +           sumsb(nv),sumsbt(nv),symdif(nv),
     +           tmpx(nv),tmpy(nv),vertex(nv),x0(nv),
     +           xdiff(nv),ydiff(nv),
     +           v(n2e),v0(n2e),
     +           velt(nmelt),
     +           vtx1(ne),vtx2(ne),w(ne),
     +           ws(ne),wsind(ne),x(nv),y(nv),
     +           xn(nv),nindex(nv),tvnS(nv),tvnSb(nv)
c     +           rcl(max2v),S(maxv),Sbar(maxv),

c     ------------------------------------------------------------------
c
c     Local scalars:
c
c          alpha0 - GRASP RCL parameter (copy)
c          faselt
c          gitr   - do loop index
c          indb
c          indw
c          j      - do loop index
c          nelt
c          objf   - value of objective function
c          randp  - portable pseudo-random number generator
c          seed0  - seed for pseudo random number generator (copy)
c          seed00 - initial seed for pseudo random number generator
c          worstz
c          zimprv
c          huge   - 2^31-1 = 2147483647
c
c     ------------------------------------------------------------------
      real       alpha0,randp
      integer    faselt,gitr,i,indb,indw,nelt,objf,objf1,objf2,seed0,
     +           seed00,worstz,zimprv,pron,poston,optz,huge
      real cput(2),btime,ttime,etime
c     ------------------------------------------------------------------
c     Adjust kmax. 
c     ------------------------------------------------------------------
      kmax=min (kmax,nv)

c     ------------------------------------------------------------------
c     Checks parameters and problem dimension.
c     ------------------------------------------------------------------
      call chkgct(errcnd,hammd,look4,maxe,maxitr,maxv,
     +            ne,nv,prttyp,seed,
     +            seedd,vtx1,vtx2)
      
      if (errcnd .gt.0) return

c     ------------------------------------------------------------------
c     Compute min diff for Elite solution update. 
c     ------------------------------------------------------------------
      mindif=ifix(hammd*nv)

c     ------------------------------------------------------------------
c     Make data structures.
c     ------------------------------------------------------------------
      call mkds (deg0,edge0,faselt,felt,first0,maxc,n2e,ne,nelt,next0,
     +           nmelt,nmxelt,nv,nxtelt,v0,vertex,vtx1,vtx2,w)

c     ------------------------------------------------------------------
c     Sort edges weights in increasing order.
c     ------------------------------------------------------------------
      print *,'into sortw ...'
      call sortw( ne,w,ws,wsind )
      print *,'outof sortw ...'

c     ------------------------------------------------------------------
c     Warm up random number generator.  Save initial seed in seed00.
c     ------------------------------------------------------------------
      seed00=seed
      call warmup(seed)

c     ------------------------------------------------------------------
c     Initialize size of largest (and smalllest elite) cut found to 
c     -inf.
c     ------------------------------------------------------------------
      huge=2147483647
      bestz=-huge
      worstz=-huge

c     ------------------------------------------------------------------
c     Copy original alpha value.
c     ------------------------------------------------------------------
      alpha0=alpha

c     ------------------------------------------------------------------
c     Print out header.
c     ------------------------------------------------------------------
      if (prttyp .ge. 1) then
           write(out,50)
50        format(//,' GRASP for Max Cut Problem-----------------',
     +           //)
          write(out,21) nv,ne,seed00
21        format(   ' input graph: ',/,
     +              ' ------vertices: ',i10,/,
     +              ' ---------edges: ',i10,//,
     +              ' seed: ',i10,//,
     +              ' iteration summary: ',/)
      endif

      
      ttime=etime(cput)
      btime=cput(1)

c     ------------------------------------------------------------------
c     Do maxitr GRASP iterations.
c     ------------------------------------------------------------------
      do 10 gitr=1,maxitr
c          print *, '***** new iteration = ',gitr
           call myflsh 
c          -------------------------------------------------------------
c          Save current seed in seed0.
c          -------------------------------------------------------------
           seed0=seed

c          -------------------------------------------------------------
c          Copy original graph onto work data structure.
c          -------------------------------------------------------------
           call cpds(deg,deg0,edge,edge0,first,first0,
     +               n2e,next,next0,
     +               nv,v,v0,vertex)

c          -------------------------------------------------------------
c          Generate random alpha if alpha is not fixed.          
c          -------------------------------------------------------------
           if (alpha0 .lt. 0) then
                alpha=randp(seed)
c               print *,'alpha = ',alpha
           endif

c          -------------------------------------------------------------
c          GRASP construction phase.  If alpha=0, then do purely
c          random construction.
c          -------------------------------------------------------------
           if (alpha .eq. 0.) then
                call buildr(nv,n2e,ne,seed,nS,S,nSbar,Sbar,vertex,objf,
     +                      first0,next0,v0,edge0,w,sums,sumsb)
	   else
                call build(alpha,degs,degsb,edge,first,maxc,
     +                 ne,n2e,n2v,next,
     +                 nS,nSbar,nv,objf,rcl,
     +                 S,Sbar,seed,sums,sumst,
     +                 sumsb,sumsbt,vertex,v,
     +                 vtx1,vtx2,w,ws,wsind)
           endif

c          print *,'S: ',(S(i),i=1,nS)
c          print *,'Sbar: ',(Sbar(i),i=1,nSbar)
           call mkcost(nv,ne,x0,nS,S,vtx1,vtx2,w,objf)
c          print *,'objf = ',objf

           objf1=objf
c          -------------------------------------------------------------
c          GRASP local search phase: VNS
c          -------------------------------------------------------------
           if (kmax.ge.2) then
                call vns(edge,first,ne,n2e,next,nS,
     +               nSbar,nv,objf,S,Sbar,sums,sumsb,
     +               v,w,x,xn,nindex,kmax,seedn,tvnS,
     +               tvnSb,look4,btime,vtx1,vtx2)
	        objf2=objf
	   else
		call local(edge,first,ne,n2e,next,nS,
     +                nSbar,nv,objf,S,Sbar,sums,sumsb,
     +                v,w)
	        objf2=objf
                
c               print *, 'after LS'
c               print *,'S: ',(S(i),i=1,nS)
c               print *,'Sbar: ',(Sbar(i),i=1,nSbar)
                call mkcost(nv,ne,x0,nS,S,vtx1,vtx2,w,objf)
c               print *,'objf = ',objf
	   endif
           if (objf.ge.look4) then
               errcnd=0
               print *,'look4 found...'
	       bestz=objf
	       call savsol(bS,bestz,bnS,gitr,gitrb,nS,nv,objf,S,
     +                            seed0,seedb,zimprv)
               ttime=etime(cput)
               print *, ' total time = ',cput(1)-btime
               return
           endif

c          print *,'after vns, objf: ', objf
c          print *,'S: ',(S(i),i=1,nS)
           call mkcost(nv,ne,x0,nS,S,vtx1,vtx2,w,objf)

c          -------------------------------------------------------------
c          If Elite set is not empty, do path-relinking from local 
c          optimal to randomly selected element of Elite set.
c          -------------------------------------------------------------
           if (pron.eq.1) then
	        if (nelt .gt. 0) then
		     if (prtype .eq. 1) then
                        call pathrl(bestz,bnS,bS,edge,faselt,felt,
     +                       first,gitr,gitrb,indb,indw,mindif,
     +                       n2e,ne,nS,
     +                       nelt,next,
     +                       nmelt,nmxelt,nv,nxtelt,objelt,objf,S,
     +                       seedp,seed0,seedb,
     +                       sums,sumsb,symdif,tmpx,tmpy,
     +                       v,velt,w,
     +                       worstz,x,y,zimprv,optS,vtx1,vtx2,x0,
     +                       xdiff,ydiff,optz,noptS)
			     objf=optz
c                     print *,' doing optS cost computation ...........'
                      print *, 'objf: ', objf, bestz
                      call mkcost(nv,ne,x0,noptS,optS,vtx1,vtx2,w,objf)
		    else
                     call prall(bestz,bnS,bS,edge,faselt,felt,
     +                  first,gitr,gitrb,indb,indw,mindif,
     +                  n2e,ne,nS,
     +                  nelt,next,
     +                  nmelt,nmxelt,nv,nxtelt,objelt,objf,S,seedp,
     +                  seed0,seedb,
     +                  sums,sumsb,symdif,tmpx,tmpy,
     +                  v,velt,w,
     +                  worstz,x,y,zimprv,optS,tmpS,xdiff,ydiff,
     +                  vtx1,vtx2)
		    endif
                else
c                    ---------------------------------------------------
c                    Elite set is empty.  Put solution in Elite set.
c                    ---------------------------------------------------
                     call updelt(bestz,bS,bnS,faselt,felt,gitr,gitrb,
     +                           indb,indw,mindif,nelt,nmelt,nmxelt,nS,
     +                           nv,nxtelt,objelt,objf,S,seed0,seedb,
     +                           xdiff,ydiff,velt,worstz,zimprv,
     +                           elitex)
		     bestz=objf
		     call savsol(bS,bestz,bnS,gitr,gitrb,nS,nv,objf,S,
     +                            seed0,seedb,zimprv)
		     zimprv=1
	        endif
	   else
		if (objf.gt.bestz) then
		      bestz=objf
		      call savsol(bS,bestz,bnS,gitr,gitrb,nS,nv,objf,S,
     +                            seed0,seedb,zimprv)
		      zimprv=1
                endif
	   endif

	   if (objf .gt. bestz) bestz=objf
c          -------------------------------------------------------------
c          If the local subroutine has found a better solution
c          -------------------------------------------------------------
           if (zimprv.eq.1) then
c             ----------------------------------------------------------
c             The output option parameter is set in order to print out 
c             the solution improvements. 
c             ----------------------------------------------------------
              if (prttyp .ge. 1) then
          	   ttime=etime(cput)
                   write(out,30) gitr,objf1,objf2,optz,bestz,
     +                           cput(1)-btime
30                 format('     itr = ',i8,
     +                    '  ph1cut = ',i10,' ***', 
     +                    '  ph2cut = ',i10,' ***', 
     +                    '     cut = ',i10,' ***', 
     +                    '    best = ',i10,' ***', 
     +                    ' time = ',f10.2)
c                  print *, 'before mkcost ...'
c        	   call myflsh
c                  call mkcost(nv,ne,x0,noptS,optS,vtx1,vtx2,w,optz)
c                  print *, 'after mkcost ...'
c                  call myflsh
              endif

c             ----------------------------------------------------------
c             The cardinality of the found cutset is still less than the
c             size of cutset sought. 
c             ----------------------------------------------------------
              if (bestz.ge.look4) then
                   errcnd=0
		   print *,'look4 found...'
                   ttime=etime(cput)
                   print *, ' total time = ',cput(1)-btime
                   return
              endif
           else
c             ----------------------------------------------------------
c             The output option parameter is set in order to print out 
c             the solution found in each GRASP iteration. 
c             ----------------------------------------------------------
              if (prttyp .eq. 2) then
                  write(out,40) gitr, objf1, objf, bestz
40                format('     itr = ',i8,' built cut = ',i6,
     +                    '     cut size = ',i6,'   best cut = ',i6)
		  call myflsh
              endif
           endif
10    continue
      if (poston.eq.1) then
            call postpr(bestz,bnS,bS,edge,faselt,felt,
     +                  first,gitr,gitrb,indb,indw,mindif,
     +                  n2e,ne,nS,
     +                  nelt,next,
     +                  nmelt,nmxelt,nv,nxtelt,objelt,objf,S,
     +                  seed0,seedb,
     +                  sums,sumsb,symdif,tmpx,tmpy,
     +                  v,velt,w,
     +                  worstz,x,y,zimprv,first0,next0,v0,edge0,
     +                  optS,xdiff,ydiff,btime)
      endif
      errcnd=0

      ttime=etime(cput)
      print *, ' total time = ',cput(1)-btime
      return
c
c     ------------------------------------------------------------------
c     End of subroutine gmcut.
c     ------------------------------------------------------------------
      end
